import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPConfiguration;
import com.telelogic.rhapsody.core.IRPPackage;
import com.telelogic.rhapsody.core.IRPProfile;
import com.telelogic.rhapsody.core.IRPProject;
import com.telelogic.rhapsody.core.IRPStereotype;
import com.telelogic.rhapsody.core.RhapsodyAppServer;

public class AndroidModelLibraryCreator {
	
	public static void main(String[] args) {
		if (args.length != 5) {
			System.out.println("Invalid arguments list: java AndroidModelLibraryCreator ANDROID_SOURCES_FOLDER ANDROID_PROFILE_NAME CREATE_ANDROID_MODEL_LIBRARY CREATE_ANDROID_STEREOTYPES CREATE_ANDROID_PROFILE");
			return;
		}
		
		// Run Example: java AndroidModelLibraryCreator L:\Android\base Android22
		ANDROID_SOURCES_FOLDER = args[0];
		ANDROID_PROFILE_NAME = args[1]; 
		CREATE_ANDROID_MODEL_LIBRARY=Boolean.parseBoolean(args[2]);
		CREATE_ANDROID_STEREOTYPES=Boolean.parseBoolean(args[3]);
		CREATE_ANDROID_PROFILE=Boolean.parseBoolean(args[4]);
		
		System.out.println(String.format("Arguments: ANDROID_SOURCES_FOLDER=%s ANDROID_PROFILE_NAME=%s CREATE_ANDROID_MODEL_LIBRARY=%b CREATE_ANDROID_STEREOTYPES=%b CREATE_ANDROID_PROFILE=%b", ANDROID_SOURCES_FOLDER, ANDROID_PROFILE_NAME, CREATE_ANDROID_MODEL_LIBRARY, CREATE_ANDROID_STEREOTYPES, CREATE_ANDROID_PROFILE));
		
		AndroidModelLibraryCreator creator = new AndroidModelLibraryCreator();
		IRPApplication app = RhapsodyAppServer.getActiveRhapsodyApplication();
		creator.setApplication(app);
		IRPProject project = app.activeProject();	
		creator.setProject(project);
		creator.execute();
	}
	
	public static String ANDROID_SOURCES_FOLDER;
	public static String ANDROID_PROFILE_NAME;
	public static boolean CREATE_ANDROID_MODEL_LIBRARY;
	public static boolean CREATE_ANDROID_STEREOTYPES;
	public static boolean CREATE_ANDROID_PROFILE;
	
	public static String ANDROID_PROFILE = "AndroidProfile";
	
	private IRPApplication application = null;
	private IRPProject project;

	public void setApplication(IRPApplication application) {
		this.application = application;
	}

	public IRPApplication getApplication() {
		return application;
	}
	
	public void setProject(IRPProject project) {
		this.project = project;
	}

	public IRPProject getProject() {
		return project;
	}	
	
	private void execute() {
		if (CREATE_ANDROID_STEREOTYPES) {
			createAndroidStereotypes();
		}
		if (CREATE_ANDROID_MODEL_LIBRARY) {
			createAndroidModelLibrary();
			setAndroidModelLibraryProperties();
		} 
		if (CREATE_ANDROID_PROFILE) {
			createAndroidProfile();
		}
	}
	
	/********************************************************************************************************
	 * Create Android Model Library
	 */
	
	private void createAndroidModelLibrary() {
		try {
			println("Creating ANDROID model library ...");
			printCurrentTime("Start Time");

			println("Project name is: " + getProject().getName());

			println("Reverse Engineer Android sources ...");
			reverseEngineerAndroidSources();
			
			println("Reverse Engineer Android sources Done.");

		} catch (Exception e) {
			println("Failed to create android model library.");
		}
	}
	
	private void reverseEngineerAndroidSources() {
		getProject().setPropertyValue("ObjectModelGe.Inheritance.line_style", "tree");
		IRPConfiguration cfg = getProject().getActiveConfiguration();
		if (cfg != null) {
			cfg.setPropertyValue("JAVA_CG.Package.GenerateDirectory", "True");
			cfg.setPropertyValue("JAVA_ReverseEngineering.ImplementationTrait.ImportAsExternal", "True");
			cfg.setPropertyValue("JAVA_ReverseEngineering.ImplementationTrait.InterfaceOnly", "True");
			cfg.setPropertyValue("JAVA_ReverseEngineering.ImplementationTrait.RootDirectory", ANDROID_SOURCES_FOLDER);
			cfg.setPropertyValue("JAVA_ReverseEngineering.Parser.IncludePath", ANDROID_SOURCES_FOLDER);
			cfg.setPropertyValue("ReverseEngineering.Main.Files",  getAndroidSourceFolders());
			cfg.setPropertyValue("ReverseEngineering.Update.CreateObjectModelDiagrams", "True");
			cfg.setPropertyValue("ReverseEngineering.Update.ObjectModelDiagramMaxElements", "100");
			cfg.setPropertyValue("ReverseEngineering.Main.UseTreeViewByDefault", "True");

			//start Reverse Engineering 
			getApplication().importClasses();
		}
	}	
	
	private String getAndroidSourceFolders() {
		List<File> files = new ArrayList<File>();
		collectAndroidSourceFiles(new File(ANDROID_SOURCES_FOLDER), files);
		Set<String> dirs = new HashSet<String>();
		for (File file : files) {
			if (isAndroidSourceFile(file)) {
				String dir = file.getParent();
				dirs.add(dir+"\\");
			}
		}
		println("Files = " + Integer.toString(files.size()));
		println("Dirs = " + Integer.toString(dirs.size()));
		String[] dirsArray = dirs.toArray(new String[dirs.size()]);
		Arrays.sort(dirsArray);
		return asList(dirsArray);
	}
	
	public boolean isAndroidSourceFile(File file) {
		boolean res = false;
		try {
			BufferedReader input =  new BufferedReader(new FileReader(file));
			try {
				String line = null;
				while (( line = input.readLine()) != null){
					if (line.startsWith("package")) {
						if (line.startsWith("package android.")) {
							res = true;
						}
						break;
					}
				}
			} finally {
				input.close();
			}
		} catch(Exception e) {
			res = false;
		}
		return res;
	}
	
	private void collectAndroidSourceFiles(File dir, List<File> files) {
		for (File file : dir.listFiles()) {
			if (file.isDirectory()) {
				collectAndroidSourceFiles(file, files);
			} else if (file.getName().endsWith(".java")) {
				files.add(file);
			}
		}
	}
	
	/********************************************************************************************************
	 * Create Android Profile
	 */
	
	private void createAndroidProfile() {
		try {
			println("Saving project ...");
			getProject().save();
			
			println("Backup previous 'Android' profile ...");
			backupPreviousAndroidProfileFolder();
			
			println("Copying 'Android' profile to Profiles folder ...");
			copyAndroidProfileToProfilesFolder();
			
			println("Creating text file ...");
			CreateTxtFile();

			printCurrentTime("End Time");
			println("Creating ANDROID profile DONE.");

		} catch (Exception e) {
			println("Failed to create 'Android' profile.");
			println(e.getMessage());
			println(getApplication().getErrorMessage());
		}
	}

	private void createAndroidStereotypes() {
		println("Creating 'Android' profile ...");
		IRPProfile profile = getProject().addProfile(ANDROID_PROFILE);
		profile.setFilename(ANDROID_PROFILE_NAME);
		if(profile != null){
			setProfileProperties(profile);
			createProfileStereotype(profile);	
		}
	}
	
	private void setProfileProperties(IRPProfile profile) {
		profile.setPropertyValue("General.Profile.AutoReferencesWithoutSubUnits", getAndroidProfilePath(false)+"\\android.sbs");
		profile.addProperty("JAVA_CG.Package.ProfileWithMainFileAnchor", "Bool", "True");
	}
	
	private void createProfileStereotype(IRPProfile profile) {
		IRPStereotype activityStereotype = addNewStereotype(profile,"AndroidActivity","Class");
		setActivityProperties(activityStereotype);

		IRPStereotype mainActivityStereotype = addNewStereotype(profile,"AndroidMainActivity","Class");
		mainActivityStereotype.addGeneralization(activityStereotype);
		setMainActivityProperties(mainActivityStereotype);

		IRPStereotype componentStereotype = addNewStereotype(profile,"AndroidComponent","Component");
		setComponentProperties(componentStereotype);
		
		IRPStereotype contentProviderStereotype = addNewStereotype(profile,"AndroidContentProvider","Class");
		setContentProviderProperties(contentProviderStereotype);

		addNewStereotype(profile,"AndroidService","Class");
		addNewStereotype(profile,"AndroidBroadcastReceiver","Class");
	}
	
	private IRPStereotype addNewStereotype(IRPProfile profile, String name, String mataClass) {
		IRPStereotype stereotype = (IRPStereotype)profile.addNewAggr("Stereotype", name);
		stereotype.addMetaClass(mataClass);
		return stereotype;
	}
	
	private void setActivityProperties(IRPStereotype activityStereotype) {
		activityStereotype.addProperty("JAVA_CG.Class.ReactiveContextArgumentNeeded", "Bool", "False");
		activityStereotype.setPropertyValue("JAVA_CG.Framework.ReactiveInit", "reactive = new Reactive();");
	}

	private void setMainActivityProperties(IRPStereotype mainActivityStereotype) {
		mainActivityStereotype.addProperty("JAVA_CG.Class.MainFileAnchor", "Bool", "True");
		mainActivityStereotype.addProperty("JAVA_CG.Framework.InitRhapsodyFrameworkInStaticBlock", "Bool", "True");
		mainActivityStereotype.addProperty("JAVA_CG.Framework.AnchorFrameworkCall", "MultiLine", "$MainFileName.InitRhapsodyFramework();");
	}
	
	private void setContentProviderProperties(IRPStereotype contentProviderStereotype) {
		contentProviderStereotype.addProperty("JAVA_CG.Framework.InitRhapsodyFrameworkInStaticBlock", "Bool", "True");
		contentProviderStereotype.addProperty("JAVA_CG.Framework.AnchorFrameworkCall", "MultiLine", "$MainFileName.InitRhapsodyFramework();");
	}
	
	private void setComponentProperties(IRPStereotype componentStereotype) {
		componentStereotype.addProperty("CG.Configuration.DestroyInitialInstance","Bool","False");
		componentStereotype.setPropertyValue("CG.Configuration.MainGenerationScheme","Full");
		componentStereotype.setPropertyValue("CG.Configuration.StartFrameworkInMainThread","False");
		componentStereotype.addProperty("JAVA_CG.Component.GenerateMainFileNearAnchorFile", "Bool", "True");
		componentStereotype.addProperty("JAVA_CG.JDK.ConvertHostToIP", "Bool", "True");
		componentStereotype.setPropertyValue("JAVA_CG.JDK.EntryPoint","onCreate");
		componentStereotype.setPropertyValue("JAVA_CG.JDK.UseRemoteHost","True");
		componentStereotype.setPropertyValue("Roundtrip.General.RoundtripScheme","Advanced");
		componentStereotype.addProperty("JAVA_CG.Framework.AnchorFrameworkCall", "MultiLine", "if (AnimInstance.isAnimationInitialized() == false) {\n    onCreate(new String[] {\"-suspendResumeAnimationByDebugger\"});\n}");
		componentStereotype.addProperty("JAVA_CG.Framework.AnchorFrameworkCallStandard", "MultiLine", "onCreate(new String[] {\"\"});");
		componentStereotype.addProperty("JAVA_CG.Framework.AnimInitMethodNameInAnchor", "String", "InitRhapsodyFramework");		
	}
	
	private void setAndroidModelLibraryProperties() {
		IRPPackage pkg = (IRPPackage)getProject().findElementsByFullName("android", "Package");
		pkg.setPropertyValue("CG.Package.UseAsExternal", "True");
		pkg.setPropertyValue("JAVA_CG.Package.DefineNameSpace", "True");
		IRPStereotype mlStereotype = (IRPStereotype)((IRPProject)pkg.getOwner()).findAllByName("ModelLibrary", "Stereotype");
		if (mlStereotype != null) {
			pkg.setStereotype(mlStereotype);
		}
	}
	
	private String getAndroidProfilePath(boolean expand) {
		String res = "$OMROOT\\Profiles\\Android\\"+ANDROID_PROFILE_NAME;
		if (expand) {
			res = res.replace("$OMROOT", getApplication().getOMROOT());
		}
		return res;
	}
	
	private void backupPreviousAndroidProfileFolder() {
		File file = new File(getAndroidProfilePath(true));
		if (file.exists()) {
			file.renameTo(new File(getAndroidProfilePath(true)+"_backup"));
		}
	}
	
	private void copyAndroidProfileToProfilesFolder() {
		try {
			// Create android profile directory
			File androidProfileDir = new File(getAndroidProfilePath(true));

			// Get directory of an android profile in a project
			String rpyPath = getProject().getCurrentDirectory() + "\\" + getProject().getName() + "_rpy";
			File rpyDir = new File(rpyPath);
			
			copyDirectory(rpyDir, androidProfileDir);
			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void CreateTxtFile() {
		String filename = getAndroidProfilePath(true)+"\\"+ANDROID_PROFILE_NAME+".txt";
		String text = String.format("This is the Rhapsody implementation of the %s profile", ANDROID_PROFILE_NAME);
		writeToFile(filename, text);
	}

	/*******************************************************************************
	 * 
	 * Utilities methods
	 */
	
	private void println(String text) {
		getApplication().writeToOutputWindow("Log", text + "\n");		
	}
	
	private void printCurrentTime(String title) {
		SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
		Date today = Calendar.getInstance().getTime();
		String time = formatter.format(today);
		println(title + ": " + time);
	}
	
	private String asList(String[] arr) {
		String s = Arrays.asList(arr).toString().replace("[", "").replace("]", "").replaceAll(" ", "").replaceAll("\t","");
		return s;
	}
	
	private void copyDirectory(File source, File destination) throws IOException {
		if (!destination.exists() && !destination.mkdirs()) {
			String message = "DirCopyFailed";
			throw new IOException(message);
		}

		for (File sourceFile : source.listFiles()) {
			if (sourceFile.isDirectory()) {
				copyDirectory(sourceFile, new File(destination, sourceFile.getName()));
			} else {
				copyFile(sourceFile, new File(destination, sourceFile.getName()));
			}
		}
	}
	
	private void copyFile (File f1, File f2) {
		InputStream in = null;
		OutputStream out = null;
		
		try{
			in = new FileInputStream(f1);
			out = new FileOutputStream(f2);
			
			byte[] buf = new byte[1024];
			int len;
			while ((len = in.read(buf)) > 0){
				out.write(buf, 0, len);
			}
		}
		catch(FileNotFoundException ex){
			println(ex.getMessage() + " in the specified directory.");
		}
		catch(IOException e){
			println(e.getMessage());      
		}
		finally {
			try {
				in.close();
				out.close();				
			} catch (IOException e) {
				println(e.getMessage());
			}	
		}
	}	
	
	private void writeToFile(String file, String text) {
		FileWriter writer;
		try {
			writer = new FileWriter(file);
			BufferedWriter out = new BufferedWriter(writer);
			out.write(text);
			out.close();
			writer.close();			
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
}
