//<

// $Id: xmlutils.inc,v 1.3 2006/12/11 15:40:42 Jerome Loubersac Exp $
// $Revision: 1.3 $
// $Date: 2006/12/11 15:40:42 $

/*
  Implements xml stream reading and writing routines
*/

/*
    Change history
    --------------
    When        Rev   Who   What
    2006-10-12  1.1   JBCL  First draft
    2006-11-29  1.2   JBCL  Added xmlWriteIfNoEmptyAttributeString routine
    2006-12-11  1.3   AJJD  Adapted for unicode

*/

//  Public Members
//      void xmlInitStream(string filename)
//      void xmlCloseStream()
//      void xmlWriteStartDocument()
//      void xmlWriteStartElement(string localName)
//      void xmlWriteEndElement(bool newLine)
//      void xmlWriteComment(string text)
//      void xmlWriteAttributeString(string localName, string value)
//      void xmlWriteIfNoEmptyAttributeString(string localName, string value)
//      void xmlWriteElementString(string localName, string value)
//      void xmlWriteIfNonEmptyElementString(string localName, string value)
//      void xmlWriteString(string value, bool newLine)
//
//      DOM_Document_ d = xmlParseFile(string xmlFileName)
//
//  Public Properties
//      bool   xmlStreamReportErrors : flag to control if errors are reported
//                  . true if errors are to be reported (default)
//                  . false for quiet mode
//
//      bool   xmlStreamErrorDialog : flag to control how reported errors are displayed
//                  . true for reporting errors with a dialog box (default)
//                  . false for reporting errors with in the DXL window


//========================================================
// Global Constants
const string xmlStreamDocStart = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"yes\"?>"
const int xmlStreamIndent = 4

const string xmlEscAmp = "&amp;"
const string xmlEscGt = "&gt;"
const string xmlEscLt = "&lt;"
const string xmlEscQuot = "&quot;"

const string msgNullxmlStream = ": null value for xmlStream"
const string msgInvalidXmlName = " is an invalid Xml name"

const Regexp rValidXmlName = regexp "^[\\.a-zA-Z0-9\\_-]*$"

//========================================================
// Global Variables
Stream xmlStream = null
int        xmlStreamIndentationLevel = 0
bool       xmlStreamStartElementOpen = false
Skip       xmlStreamStack = null Skip
int        xmlStreamElementCounter = 0
bool       xmlDocumentEmpty = true
bool       xmlDocumentHasRoot = false
int        xmlStreamErrorCounter = 0

Buffer xmlBuff = null Buffer


//========================================================
// Global Properties
bool   xmlStreamReportErrors = true    // flag to control if errors are reported
bool   xmlStreamErrorDialog  = false   // flag to control how reported errors are displayed


//========================================================
// Private Functions and Routines


Regexp hasCodepoint = regexp "[\\\\]u([0-9]+)[?]"


//========================================================
// Function Name: convertToCodePoints
//
// Description:   Convert all non-ascii characters into the RTF unicode representation
//                in the form \u12345?
//
// Type: Public
//
// Precondition:   None
//
void convertToCodePoints(Buffer b)
{
    if ( null b || length(b) == 0 ) return

    Buffer r = create()
    int i
    for i in 0:length(b)-1 do
    {
        if ( !isascii(b[i]) )
        {
            r += "\\u"
            r += codepointOf(b[i])""
            r += "?"
        }
        else
        {
            combine(r, b, i, i)
        }
    }

    setempty(b)
    combine(b, r, 0)
    delete(r)
}


//========================================================
// Function Name: convertFromCodePoints
//
// Description:   Convert all RTF unicode representations of the form \u12345?
//                back into Unicode characters
//
// Type: Public
//
// Precondition:   None
//
void convertFromCodePoints(Buffer b)
{
    if ( null b || length(b) == 0 ) return

    Buffer r = create()
    int i = 0
    while ( search(hasCodepoint, b, i) )
    {
        int ch = intOf(b[i+start(1):i+end(1)])
        combine(r, b, i, i+start(0)-1)
        r += wcharOf(ch)

        i += end(0)+1
    }

    combine(r, b, i)

    setempty(b)
    combine(b, r, 0)
    delete(r)
}


//========================================================
// Function Name: parseXMLBufferUnicode
//
// Description:   Overloading of parseXMLBuffer_ for Unicode treatment
//
// Type: Public
//
// Precondition:   None
//
DOM_Document_ parseXMLBufferUnicode(Buffer b)
{
    return parseXMLBuffer_(b)
}


DOM_Document_ parseXMLBuffer_(Buffer b)
{
    convertToCodePoints(b)
    return parseXMLBufferUnicode(b)
}


//========================================================
// Function Name: parseXMLString_
//
// Description:   Overloading of parseXMLString_ for Unicode treatment
//
// Type: Public
//
// Precondition:   None
//
DOM_Document_ parseXMLString_(string s)
{
    Buffer b = create()
    b += s
    convertToCodePoints(b)
    DOM_Document_ d = parseXMLBufferUnicode(b)
    delete(b)

    return d
}


//========================================================
// Function Name: getStringAttributeUnicode
//
// Description:   Overloading of getStringAttribute_ for Unicode treatment
//
// Type: Public
//
// Precondition:   None
//
string getStringAttributeUnicode(DOM_Element_ d, string s)
{
    return getStringAttribute_(d, s)
}


string getStringAttribute_(DOM_Element_ d, string s)
{
    if ( null s ) return ""

    string r = getStringAttributeUnicode(d, s)

    if ( !null r )
    {
        Buffer b = create()
        b += r
        convertFromCodePoints(b)
        r = stringOf(b)
        delete(b)
    }

    return r
}


//========================================================
// Function Name: getElementTextUnicode
//
// Description:   Overloading of getElementText_ for Unicode treatment
//
// Type: Public
//
// Precondition:   None
//
string getElementTextUnicode(DOM_Element_ d)
{
    return getElementText_(d)
}


string getElementText_(DOM_Element_ d)
{
    Buffer b = create()
    b += getElementTextUnicode(d)

    convertFromCodePoints(b)

    string r = stringOf(b)
    delete(b)

    return r
}



//========================================================
// Function Name: xmlStreamIndentation
//
// Description: Generates and return an indentation string
//              used to prefix lines output to the xmlStream
//
//
// Type: Private
//
// Precondition: none
//
string xmlStreamIndentation()
{
    string result = ""

    int i = xmlStreamIndentationLevel * xmlStreamIndent

    Buffer b = create(i)

    while (i > 0)
    {
        b += " "
        i = i - 1
    }

    result = stringOf(b)
    delete(b)

    return result
}


//========================================================
// Function Name: xmlEscapeSpecialCharacters
//
// Description: Returns a version of the input string where
//              special characters " & < and > have been replaced
//              by their ampersand escaped equivalent defined
//              in the constants xmlEscQuot, xmlEscAmp, xmlEscLt
//              and xmlEscGt
//
// Type: Private
//
// Precondition: none
//
string xmlEscapeSpecialCharacters(string text)
{
    string result = ""

    Buffer bSource = create()
    Buffer bTarget = create()

    if (!null text)
    {


        bSource += text

        int i

        for i in 0:length(bSource)-1 do
        {
            if ( bSource[i] == '<' )
            {
                bTarget += xmlEscLt
            }
            else if ( bSource[i] == '>' )
            {
                bTarget += xmlEscGt
            }
            else if ( bSource[i] == '&' )
            {

                bTarget += xmlEscAmp
            }

            else if ( bSource[i] == '\"' )
            {
                bTarget += xmlEscQuot

            }
            else if( bSource[i] == '\'' )
       		{
	       		bTarget += "&apos;"
	    	}
            else
            {

                bTarget += bSource[i]""
            }
        }

        result = stringOf(bTarget)
    }

    delete(bSource)

    delete(bTarget)

    return result
}


//========================================================
// Function Name: xmlStreamError
//
// Description: Displays the input message as an error report.
//
//              The properties xmlStreamReportErrors and
//              xmlStreamErrorDialog control the error
//              reporting modes
//
// Type: Private
//
// Precondition: none
//
void xmlStreamError(string message)
{
    // count errors regardless of whether they are reported or not
    xmlStreamErrorCounter = xmlStreamErrorCounter + 1;

    // Report the error unless quiet mode is selected
    if (xmlStreamReportErrors)
    {
        // add some information to the input message
        string strMessage = "*** xmlStream error - " message "\n"

        // report an error according to selected reporting mode
        if (xmlStreamErrorDialog)
        {
            errorBox strMessage
        }
        else
        {
            print strMessage
        }
    }
}


//========================================================
// Function Name: xmlStreamAppend
//
// Description: Appends the input line to the currently open
//              xmlStream
//
// Type: Private
//
// Precondition: none
//
void xmlStreamAppend(string strLine)
{
    if (! null xmlStream)
    {
        xmlStream << strLine
    }
    else
    {
        xmlStreamError("Error - cannot write to xmlStream")
    }
}


//========================================================
// Function Name: xmlStreamAppendIndented
//
// Description: Appends the input line to the currently open
//              xmlStream, at the current indentation level
//
// Type: Private
//
// Precondition: none
//
void xmlStreamAppendIndented(string strLine)
{
    xmlStreamAppend(xmlStreamIndentation)
    xmlStreamAppend(strLine)
}


//========================================================
// Function Name: xmlStreamIsOpen
//
// Description: Returns true if the xmlStream is open, otherwise
//              returns false
//
// Type: Private
//
// Precondition: none
//
bool xmlStreamIsOpen()
{
    return (! null xmlStream)
}


//========================================================
// Public Members


//========================================================
// Function Name: xmlInitStream
//
// Description: Initialise the xmlStream and open a stream
//              of the given filename.
//
// Type: Public
//
// Precondition: none
//
void xmlInitStream(string filename)
{
    // create a new stream
    xmlStream = write (filename, CP_UTF8)
    xmlBuff = create()

    // initialise the xmlStream elements
    xmlStreamIndentationLevel = 0
    xmlStreamStartElementOpen = false
    xmlStreamStack = create()
    xmlStreamElementCounter = 0
    xmlDocumentEmpty = true
    xmlStreamErrorCounter = 0
    xmlDocumentEmpty = true
    xmlDocumentHasRoot = false
}


//========================================================
// Function Name: xmlCloseStream
//
// Description: Flushes and closes the xmlStream stream
//              and resets the xmlStream itself
//
// Type: Public
//
// Precondition: none
//
void xmlCloseStream()
{
    // check that the xmlStream has been created
    if (!null xmlStream)
    {
        // if some elements are still open...
        if (xmlStreamElementCounter > 0)
        {
            //... report an error
            xmlStreamError("unexpected ClosexmlStream()" )
        }
        else
        {
            //... otherwise flush and close the stream
//xmlStream << tempStringOf(xmlBuff)
delete(xmlBuff)

//	        flush xmlStream
            close xmlStream

            // ... and reset all the xmlStream elements
            // (but except the error counter so it can be available)
            xmlStream = null
            xmlDocumentEmpty = true
            xmlStreamElementCounter = 0
            delete(xmlStreamStack)
            flushDeletions
            xmlStreamStack = null Skip
            xmlStreamStartElementOpen = false
            xmlStreamIndentationLevel = 0
            xmlStreamErrorCounter = 0
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected ClosexmlStream() stream" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteStartDocument
//
// Description: Writes the standard xml declaration defined
//              by the global constant xmlStreamDocStart.
//
//              And Xml document does not have to have a declaration
//              but if it has, it must be the first thing
//              in the document.
//
// Type: Public
//
// Precondition: none
//
void xmlWriteStartDocument()
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // ... if it is not the first thing in the document
        if (!xmlDocumentEmpty)
        {
            // ... report an error
            xmlStreamError("unexpected WriteStartDocument()" )
        }
        else
        {
            // ... otherwise output the standard document start to the xmlStream stream
            xmlStreamAppend(xmlStreamDocStart)
            xmlStreamAppend("\n")

            // ... and indicate that the document is no longer empty
            xmlDocumentEmpty = false
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteStartDocument()" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteStartElement
//
// Description: Writes the start-tag for the given element.
//              The closing > is deferred to allow attributes to be written
//              inside the tag.
//              Stores each element on a stack to allow the corresponding WriteEndElement
//
// Type: Public
//
// Precondition: none
//
void xmlWriteStartElement(string localName)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // ensure that the document only has one root element
        if ((xmlStreamIndentationLevel == 0) && (xmlDocumentHasRoot))
        {
            // report the invalid xml name error
            xmlStreamError("unexpected WriteStartElement(\"" localName "\")")
        }
        else
        {
            // if a previous start-tag had its closing > deferred ...
            if (xmlStreamStartElementOpen)
            {
                //... it can be now be closed
                xmlStreamAppend(">\n")
                xmlStreamStartElementOpen = false
            }

            // check that the localName does not contain illegal character
            if (rValidXmlName localName)
            {
                // ... write the start-tag (deferring the closing > so that attributes may be added)
                xmlStreamAppendIndented("<")
                xmlStreamAppend(localName)

                xmlStreamStartElementOpen = true

                // ... add local name to stack
                xmlStreamElementCounter = xmlStreamElementCounter + 1
                put(xmlStreamStack, xmlStreamElementCounter, localName)

                // ... increment the indentation level
                xmlStreamIndentationLevel = xmlStreamIndentationLevel + 1

                // ... and indicate that the document is no longer empty
                xmlDocumentEmpty = false
                xmlDocumentHasRoot = true
           }
           else
            {
                // report the invalid xml name error
                xmlStreamError("WriteStartElement(\"" localName "\") : "  localName msgInvalidXmlName "")
            }
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteStartElement(\"" localName "\")" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteEndElement
//
// Description: Writes the end-tag for innermost opened WriteStartElement.
//              Find the appropriate element name then remove it from the stack.
////
// Type: Public
//
// Precondition: none
//
void xmlWriteEndElement(bool newLine)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        if (xmlStreamElementCounter > 0)
        {
            // decrease indent
            xmlStreamIndentationLevel = xmlStreamIndentationLevel - 1

            if (xmlStreamStartElementOpen)
            {
                // the start tag may now be closed and this is necessarily
                // an empty element
                xmlStreamAppend("/>\n")
                xmlStreamStartElementOpen = false
            }
           else
            {
                // get local name from stack
                string strTag = ""
                find(xmlStreamStack, xmlStreamElementCounter, strTag)

                if(newLine)
                {
	                // write a new line write
	                //xmlStreamAppend("\n")
	                // then the corresponding end element, indented
                    xmlStreamAppendIndented("</")
                    xmlStreamAppend(strTag)
                    xmlStreamAppend(">\n")
                }
                else
                {
                    // write the corresponding end element
                    xmlStreamAppend("</")
                    xmlStreamAppend(strTag)
                    xmlStreamAppend(">\n")
                }
            }

            // in all cases, remove local name from stack
            delete(xmlStreamStack, xmlStreamElementCounter)
            flushDeletions

            // and decrease the element counter
            xmlStreamElementCounter = xmlStreamElementCounter - 1
        }
        else
        {
            // no open element, so report an error
            xmlStreamError("unexpected WriteEndElement()")
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteEndElement()" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteComment
//
// Description: Writes the input text as Xml comment
//
// Type: Public
//
// Precondition: none
//
void xmlWriteComment(string text)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // write the text inside some comment start and end tags
        string comment =  text "-->"
        xmlStreamAppendIndented("<!--")
        xmlStreamAppend(comment)
        xmlStreamAppend("\n")

        // indicate that the document is no longer empty
        xmlDocumentEmpty = false
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteComment(\"" text "\")" msgNullxmlStream)
    }
}

//========================================================
// Function Name: xmlWriteAttributeString
//
// Description: Writes the attribute and value inside the currenty open start-tag.
//              Start-tags are left open by deferring the writing of the closing >.
//
// Type: Public
//
// Precondition: none
//
void xmlWriteAttributeString(string localName, string value)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // A start-tag must be currently open
        if (xmlStreamStartElementOpen)
        {
            // check that the localName does not contain illegal character
            //if (isValidXmlName(localName))
            if (rValidXmlName localName)
            {
                //string strValue = xmlEscapeSpecialCharacters(value)
                Buffer b=create
                b=xmlEscapeSpecialCharacters(value)

                // ... write the attribute name and value
                xmlStreamAppend(" ")
                xmlStreamAppend(localName)
                xmlStreamAppend("=\"")
                xmlStreamAppend(stringOf(b))
                xmlStreamAppend("\"")
                
                delete(b)
            }
            else
            {
                // report the invalid xml name error
                xmlStreamError("WriteStartElement(\"" localName "\") : "  localName msgInvalidXmlName "")
            }
        }
        else
        {
            // ... otherwise report an error
            xmlStreamError("unexpected WriteAttributeString(\"" localName "\", \"" value "\")")
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteAttributeString(\"" localName "\", \"" value "\")" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteIfNonEmptyAttributeString
//
// Description: Routine to write an attribute
//                to the the xmlStream only if its
//                value is non empty, otherwise ignore it.
//
// Type: Public
//
// Precondition: none
//
void xmlWriteIfNonEmptyAttributeString(string localName, string value)
{
    if (value != "")
    {
        xmlWriteAttributeString(localName, value)
    }
}

//========================================================
// Function Name: xmlWriteElementString
//
// Description:  Writes an element of localName containing the
//               specified value.
//
// Type: Public
//
// Precondition: none
//
void xmlWriteElementString(string localName, string value)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // ensure that the document only has one root element
        if ((xmlStreamIndentationLevel == 0) && (xmlDocumentHasRoot))
        {
            // report the invalid xml name error
            xmlStreamError("unexpected WriteStartElement(\"" localName "\")")
        }
        else
        {
            // if a previous start-tag had its closing > deferred ...
            if (xmlStreamStartElementOpen)
            {
                //... it can be now be closed
                xmlStreamAppend(">\n")
                xmlStreamStartElementOpen = false
            }

            // check that the localName does not contain illegal character
            //if (isValidXmlName(localName))
            if (rValidXmlName localName)
            {
                string strValue = xmlEscapeSpecialCharacters(value)

                // the end tag is written immediately, so no need to put the element on the stack
                xmlStreamAppendIndented("<")
                xmlStreamAppend(localName)
                xmlStreamAppend(">")
                xmlStreamAppend(strValue)
                xmlStreamAppend("</")
                xmlStreamAppend(localName)
                xmlStreamAppend(">\n")

                // indicate that the document is no longer empty
                xmlDocumentEmpty = false
                xmlDocumentHasRoot = true
            }
            else
            {
                // report the invalid xml name error
                xmlStreamError("WriteElementString(\"" localName "\", \"" value "\") : "  localName msgInvalidXmlName "")
            }
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteElementString(\"" localName "\", \"" value "\")" msgNullxmlStream)
    }
}


//========================================================
// Function Name: xmlWriteIfNonEmptyElementString
//
// Description:   Routine to write an element
//                to the the xmlStream only if its
//                value is non empty, otherwise ignore it.
//
// Type: Private
//
// Precondition: none
//
void xmlWriteIfNonEmptyElementString(string sName, string sValue)
{
    if (sValue != "")
    {
        xmlWriteElementString(sName, sValue)
    }
}


//========================================================
// Function Name: confFileContents
//
// Description:   Read the contents of a file into a buffer of adequate size.
//
// Type: Public
//
// Precondition:
//
Buffer confFileContents(string xmlFileName, ConfType cType)
{
    int    fileSize = 10000
    Buffer xbuffer  = create(fileSize)

    if ( confFileExists(xmlFileName, cType) )
    {
        // Open the file as stream
        ConfStream cs = confRead(xmlFileName, cType)

        // Copy all its content in a buffer so that it can be parsed
        cs >> xbuffer

        // ... gradually increase size of buffer until the whiole file is read
        while ( length(xbuffer) > fileSize-4 )
        {
            fileSize += 10000
            delete(xbuffer)
            xbuffer = create(fileSize)
            close(cs)
            cs = confRead(xmlFileName, cType)
            cs >> xbuffer
        }

        // ... and release the stream
        close(cs)
    }

    return xbuffer
}


//========================================================
// Function Name: xmlWriteString
//
// Description:   Routine to write the supplied string to
//                the xmlStream. The boolean parameter
//                newLine controls whether the string should be
//                on a new line or not.
//
// Type: Public
//
// Precondition: none
//
void xmlWriteString(string value, bool newLine)
{
    // check that the xmlStream has been created
    if ( xmlStreamIsOpen )
    {
        // check if we are inside an element
        if (xmlStreamElementCounter > 0)
        {
            // escape special characters
            //string strValue = xmlEscapeSpecialCharacters(value)

            // if a previous start-tag had its closing > deferred ...
            if (xmlStreamStartElementOpen)
            {
                //... it can be now be closed
                xmlStreamAppend(">")
                if(newLine)
                {
                    xmlStreamAppend("\n")
                }
                xmlStreamStartElementOpen = false
            }

            if(newLine)
            {
                // write the escaped text at the curent identation level
                //xmlStreamAppendIndented(strValue)
                //write whole value as it comes
                xmlStreamAppendIndented(value)
                xmlStreamAppend("\n")
            }
            else
            {
                //xmlStreamAppend(strValue)
                //write whole value as it comes
                xmlStreamAppend(value)
            }
        }
        else
        {
            //... otherwise report an error
            xmlStreamError("unexpected WriteString(\"" value "\")")
        }
    }
    else
    {
        // report the undefined xmlStream error
        xmlStreamError("unexpected WriteString(\"" value "\")" msgNullxmlStream)
    }
}


DOM_Document_ xmlParseConfFile(string xmlFileName, ConfType cType)
{
    DOM_Document_ result = null

    if(confFileExists(xmlFileName, cType))
    {
        // Get contents of file into buffer
        Buffer xbuffer = confFileContents(xmlFileName, cType)

        // ... and finally parse the buffer into an XML document
        result = parseXMLBuffer_(xbuffer)

        // ... when done, release the buffer
        delete xbuffer
    }

    return result

}


string tempConfFileName(ConfType cType, string ext)
{
    string tmpFileName = ""

    string fileRoot = "TMP"

    int i = 0
    for i in 0:999 do
    {
        tmpFileName = fileRoot i ext

        if ( !confFileExists(tmpFileName, cType) )
        {
            break
        }
    }

    return tmpFileName
}


//========================================================
// Function Name: xmlParseFile
//
// Description:   Parse the specified file and returns the
//                corresponding DOM_Document_.
//
// Type: Public
//
// Precondition: none
//
DOM_Document_ xmlParseFile(string xmlFileName)
{
    DOM_Document_ result = null

    if(fileExists(xmlFileName))
    {
        string tempXmlFile = tempConfFileName(confSystem, ".xml")

        string message = confUploadFile(xmlFileName, tempXmlFile, confSystem)

        result = xmlParseConfFile(tempXmlFile, confSystem)
        
        safelyDeleteConfFile(tempXmlFile, confSystem)
    }

    return result
}


